# Copyright (c) 2021, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Class describing a database, working with a database engine."""

from typing import Optional, Sequence, Type, Union

from mustang.engine.base import BaseEngine
from mustang.engine.sqlite import Sqlite3Engine
from mustang.schema.model import Model, MODELS

class Database:

    """
    A Mustang database, linking the ORM with an engine.

    This is the main class for all databases.  It is linked to a database
    engine (sqlite by default) and will delegate to this engine most
    of the tasks it should perform.  However, this class also provides
    the true meat of the Object-Relational Mapping (ROM).

    It is possible to ovverride this class to change most aspects of the
    library, optimize some details, alter some behavior.

    """

    def __init__(self):
        self._engine = Sqlite3Engine(self)
        self._models = ()

    @property
    def engine(self):
        """Return the database engine currently being used."""
        return self._engine
    @engine.setter
    def engine(self, engine: Union[BaseEngine, Type[BaseEngine]]):
        """
        Change the engine in use.

        Args:
            engine (Engine or subclass of Engine): the new database engine.
                    It can be instantiated or not.  It is preferrable
                    to let this property handles instantiation.

        """
        if issubclass(engine, BaseEngine):
            engine = engine(self)
        else:
            engine.database = self

        self._engine = engine

    def bind(self, models: Optional[Sequence[Model]] = None):
        """
        Bind this database to the specified, or the detected, models.

        If this method is called with no argument, all imported classes
        inherited from models are bound with this database.  You can
        also spedcify a list (or a sequence of any type) of models to bind.

        Args:
            models (optional, sequence of Model): the models to bind.

        """
        models = MODELS if models is None else models

        # Check that all models equire no external bound models.
        for cls in models:
            # ... add checks here.
            cls.load_schema()
            cls._engine = self._engine

        self.models = tuple(models)

    def init(self, *args, **kwargs):
        """
        Initialize (create if necessary) the database.

        Positional or keyword arguments are sent to the database engine.

        """
        self._engine.init(*args, **kwargs)
        self._engine.create_migration_table()

        # Check the model schemas.
        for cls in self.models:
            saved_schema = self._engine.get_saved_schema_for(cls)
            if saved_schema is None:
                self._engine.create_table_for(cls)
            else:
                diff = self.create_diff_between(saved_schema, cls._schema)
                if diff is not None:
                    self._engine.apply_diff_for(cls, diff)

    def close(self):
        """Close the database."""
        return self._engine.close()

    def destroy(self):
        """Destroy the database."""
        self._engine.destroy()



# Copyright (c) 2021, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Module defining the model's schema."""

from enum import Enum

from mustang.exceptions import ForbiddenArgument, MissingArgument

_NOT_SET = object()

class Operation(Enum):

    """Operation type for schema editing/extracting."""

    CREATION = "creation"
    PORTION = "portion"


class ModelSchema:

    """
    Class to represent the model's current (and past) schema.
    """

    def __init__(self, model, fields):
        self.model = model
        self.fields = fields
        self.primary_key = next(field for field in fields.values()
                if field.primary_key)

    def extract(self, args, kwargs, operation: Operation = Operation.PORTION):
        """Extract and return a dictionary of {key: value} for this schema."""
        fields = iter(self.fields.values())
        for arg in args:
            field = None
            while field is None and (not default or field.set_by_database):
                try:
                    field = next(fields)
                except StopIteration:
                    raise ValueError(
                            f"field {arg!r}: cannot find a matching "
                            "field.  It might be better to use "
                            "keyword arguments."
                    )

            key = field.name
            if key in kwargs:
                raise ValueError(
                        f"argument {name!r} has been defined both with "
                        "a positional and keyword argument.   It might "
                        "be good to use only keyword arguments to avoid "
                        "this confusion"
                )
            kwargs[key] = arg

        # Now check kwargs
        fields = {}
        for field in self.fields.values():
            value = kwargs.get(field.name, _NOT_SET)
            if value is _NOT_SET:
                if operation is Operation.PORTION:
                    continue

                if field.has_default:
                    value = field.default
                elif field.set_by_database:
                    continue
                else:
                    raise MissingArgument(self.model, field)
            elif operation is Operation.CREATION and field.set_by_database:
                raise ForbiddenArgument(self.model, field)

            fields[field] = value

        return fields



# Copyright (c) 2021, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Module containing the Sqlite3 database engine."""

import datetime
import pathlib
import pickle
import sqlite3
from textwrap import dedent
from typing import Any, Dict, Optional, Type, Union

from mustang.engine.base import BaseEngine
from mustang.schema.field import Field

Model = 'mustang.schema.model.Model'

class Sqlite3Engine(BaseEngine):

    """
    The sqlite3 dabase engine for Mustang.

    This engine allows to connect and query a Sqlite3 database (stored
    in a file or in memory).

    """

    def __init__(self, database):
        super().__init__(database)
        self.file_name = None
        self.memory = False

    def init(self, file_name: Union[str, pathlib.Path, None] = None,
            memory: bool = False):
        """
        Initialize the database engine.

        Args:
            file_name (str or Path): the file_name in which the database
                    is stored, or will be stored.  It can be a relative
                    or absolute file name.  If you want to just store
                    in memory, don't specify a file name and just set
                    the `memory` argument to `True`.
            memory (bool): whether to store this database in memory or
                    not?  If `True`, the file name is ignored.

        """
        self.file_name = file_name if not memory else None
        self.memory = memory

        # Connect to the database.
        if memory:
            sql_file_name = ":memory:"
        else:
            if isinstance(file_name, str):
                file_name = pathlib.Path(file_name)

            assert isinstance(file_name, pathlib.Path)
            if file_name.absolute():
                sql_file_name = str(file_name)
            else:
                sql_file_name = str(file_name.resolve())
            self.file_name = file_name
        self.connection = sqlite3.connect(sql_file_name)
        self.cursor = self.connection.cursor()

    def close(self):
        """Close the database."""
        if self.connection:
            self.connection.close()

    def destroy(self):
        """Destroy the database."""
        self.close()
        if self.file_name:
            self.file_name.unlink()

    def create_migration_table(self):
        """
        Create the migration table, if it doesn't exist.

        Notice that this method will be called each time the engine
        is created, so this method should do nothing if the migration
        table already exists.

        """
        self.cursor.execute(CREATE_MIGRATION_TABLE)

    def create_table_for(self, model: Type[Model]):
        """
        Create a table for this model.

        Notice that this method is called each time the model is
        bound, therefore this method should do nothing if the table
        already exists.

        Args:
            model (subclass of Model): the model.

        """
        table_name = model._alt_name or model.__name__.lower()
        sql_fields = []
        for field in model._fields.values():
            sql_type = SQL_TYPES.get(field.field_type, "BLOB")
            sql_field = f"{field.name} {sql_type}"
            if field.primary_key:
                sql_field += " PRIMARY KEY"
                if field.field_type is int:
                    # Autoincrement by default on primary key ints.
                    sql_field += " AUTOINCREMENT"

            if not field.has_default:
                sql_field += " NOT NULL"

            sql_fields.append(sql_field)

        # Send the create query.
        fields = ", ".join(sql_fields)
        self.cursor.execute(CREATE_TABLE.format(table_name=table_name,
                fields=fields))

    def get_saved_schema_for(self, model: Type[Model]):
        """
        Return the saved schema for this model, if any.

        Returning `None` will lead to the database calling `create_table_for`.
        If migrations are supported for this engine, the schema should
        be returned (the list of fields stored in the last migration).

        Args:
            model (subclass of Model): the model.

        """
        return None

    def get_instance(self, model: Type[Model],
            fields: Dict[Field, Any]) -> Optional[Model]:
        """
        Get, if possible, an instance with the specified fields.

        If more than one instance would match the specified fields,
        `None` is returned.  If no match is found, `None` is also returned.
        For greater precision, use `select`.

        Args:
            model (subclass of Model): the model class.
            fields (dict): the field dictionary, containing, as keys,
                    field objects, ans as values, whatever value
                    (of whatever type) has been set by the user.

        Returns:
            instance (Model or None): the instance matching these fields.

        """
        table_name = model._alt_name or model.__name__.lower()
        sql_fields = []
        sql_values = []
        for field, value in fields.items():
            sql_field = f"{field.name}=?"
            sql_fields.append(sql_field)
            sql_values.append(value)

        # Determine the field of the queries.
        fields = []
        for field in model._fields.values():
            fields.append(field.name)
        fields = ", ".join(fields)

        # Send the query.
        filters = " AND ".join(sql_fields)
        rows = self._execute(SELECT_QUERY.format(table_name=table_name,
                fields=fields, filters=filters), sql_values)

        # Loop over the rows.
        rows = self.cursor.fetchall()
        if len(rows) == 0 or len(rows) < 1:
            return None

        row = rows[0]
        instance_data = {}
        for i, field in enumerate(model._fields.values()):
            value = row[i]

            # Convert this SQL value to Python if necessary.
            if isinstance(value, bytes) and field.field_type is not bytes:
                # Unpickle the data.
                value = pickle.loads(value)

            instance_data[field.name] = value

        # Create an instance object.
        return model(**instance_data)

    def create_instance(self, model: Type[Model], fields: Dict[Field, Any]):
        """
        Create and update a model's instance fields.

        Args:
            instance (Model): the model to be populated.
            fields (dict): the dictionary or fields.  This should contain
                    field objects as keys and their values (can be
                    a default value).

        """
        table_name = model._alt_name or model.__name__.lower()
        sql_fields = []
        sql_values = []
        for field, value in fields.items():
            sql_fields.append(field.name)
            sql_values.append(value)

        # Send the query.
        values = ", ".join(["?"] * len(fields))
        sql_fields = ", ".join(sql_fields)
        self._execute(INSERT_QUERY.format(table_name=table_name,
                fields=sql_fields, values=values), sql_values)
        instance_data = {}
        for field in model._fields.values():
            value = fields.get(field)
            if field.set_by_database:
                value = self.cursor.lastrowid

            instance_data[field.name] = value

        # Create an instance object.
        return model(**instance_data)

    def update_instance(self, instance: Model, field: Field, value: Any):
        """
        If possible, update the specific instance's field.

        Args:
            instance (Model): the model to modify.
            field (Field): the field to be applied.
            value (Any): the field's new value.

        This value is supposed to have been filtered and allowed by the
        instance.

        """
        # Get the primary key field.
        model = type(instance)
        table_name = model._alt_name or model.__name__.lower()
        primary = instance._schema.primary_key
        id_value = getattr(instance, primary.name)
        sql_values = (value, id_value)
        self._execute(UPDATE_QUERY.format(table_name=table_name,
                primary=primary.name, field=field.name), sql_values)
        instance._has_init = False
        setattr(instance, field.name, value)
        instance._has_init = True

    def delete_instance(self, instance: Model):
        """
        Delete the specified instance from the database.

        Args:
            instance (Model): the model instance to be deleted.

        """
        # Get the primary key field.
        model = type(instance)
        table_name = model._alt_name or model.__name__.lower()
        primary = instance._schema.primary_key
        id_value = getattr(instance, primary.name)
        self._execute(DELETE_QUERY.format(table_name=table_name,
                primary=primary.name), (id_value, ))
        instance._is_deleted = True

    def _execute(self, query, fields=None):
        """Execute a query."""
        fields = fields if fields is not None else ()
        return self.cursor.execute(query, fields)


## Constants
SQL_TYPES = {
        int: "INTEGER",
        float: "REAL",
        str: "TEXT",
        bytes: "BLOB",
        datetime.datetime: "TIMESTAMP",
        datetime.date: "DATE",
}

# Database queries
CREATE_MIGRATION_TABLE = dedent("""
    CREATE TABLE IF NOT EXISTS mustang_migration (
        id INTEGER PRIMARY KEY AUTOINCREMENT,
        table_name TEXT UNIQUE NOT NULL,
        last_updated TIMESTAMP NOT NULL,
        schema BLOB NOT NULL
    );
""".strip("\n"))

CREATE_TABLE = dedent("""
    CREATE TABLE IF NOT EXISTS {table_name} (
        {fields}
    );
""".strip("\n"))

SELECT_QUERY = dedent("""
    SELECT {fields} FROM {table_name}
    WHERE {filters};
""".strip("\n"))

INSERT_QUERY = dedent("""
    INSERT INTO {table_name} ({fields})
    VALUES ({values});
""".strip("\n"))

UPDATE_QUERY = dedent("""
    UPDATE {table_name}
    SET {field}=?
    WHERE {primary}=?
""".strip("\n"))

DELETE_QUERY = dedent("""
    DELETE FROM {table_name}
    WHERE {primary}=?
""".strip("\n"))



# Copyright (c) 2021, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Module containing the base class for a database engine."""

from abc import ABCMeta, abstractmethod
from typing import Any, Dict, Optional, Type

from mustang.schema.field import Field

Model = 'mustang.schema.model.Model'

class BaseEngine(metaclass=ABCMeta):

    """
    Base class for a database engine.

    Adding a database engine can be done by inheriting from `BaseEngine`
    and setting the new engine in the database.

    """

    def __init__(self, database):
        self.database = database

    @abstractmethod
    def init(self, *args, **kwargs):
        """
        Initialize the database engine.

        Override this method in sub-classes.

        Optional and keyword arguments are supported.

        """

    @abstractmethod
    def close(self):
        """Close the database."""

    @abstractmethod
    def destroy(self):
        """Destroy the database."""

    @abstractmethod
    def create_migration_table(self):
        """
        Create the migration table, if it doesn't exist.

        Notice that this method will be called each time the engine
        is created, so this method should do nothing if the migration
        table already exists.

        """

    @abstractmethod
    def create_table_for(self, model):
        """
        Create a table for this model.

        Notice that this method is called each time the model is
        loaded, therefore this method should do nothing if the table
        already exists.

        Args:
            model (subclass of Model): the model.

        """

    @abstractmethod
    def get_saved_schema_for(self, model):
        """
        Return the saved schema for this model, if any.

        Returning `None` will lead to the database calling `create_table_for`.
        If migrations are supported for this engine, the schema should
        be returned (the list of fields stored in the last migration).

        Args:
            model (subclass of Model): the model.

        """

    @abstractmethod
    def get_instance(self, model: Type[Model],
            fields: Dict[Field, Any]) -> Optional[Model]:
        """
        Get, if possible, an instance with the specified fields.

        If more than one instance would match the specified fields, `None` is returned.  If no match is found, `None` is also returned.  For greater precision, use `select`.

        Args:
            model (subclass of Model): the model class.
            fields (dict): the field dictionary, containing, as keys,
                    field objects, ans as values, whatever value
                    (of whatever type) has been sent by the user.

        Returns:
            instance (Model or None): the instance matching these fields.

        """


    @abstractmethod
    def create_instance(self, instance: Model, fields: Dict[Field, Any]):
        """
        Create and update a model's instance fields.

        Args:
            instance (Model): the model to be populated.
            fields (dict): the dictionary or fields.  This should contain
                    field objects as keys and their values (can be
                    a default value).

        """

    @abstractmethod
    def update_instance(self, instance: Model, field: Field, value: Any):
        """
        If possible, update the specific instance's field.

        Args:
            instance (Model): the model to modify.
            field (Field): the field to be applied.
            value (Any): the field's new value.

        This value is supposed to have been filtered and allowed by the
        instance.

        """

    @abstractmethod
    def delete_instance(self, instance: Model):
        """
        Delete the specified instance from the database.

        Args:
            instance (Model): the model instance to be deleted.

        """



# Copyright (c) 2021, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Module containing the field class, defining database columns."""

from typing import Any

_NOT_SET = object()

class Field:

    """A field, to represent a database column."""

    def __init__(self, field_type, primary_key=False,
            name=None, default=_NOT_SET):
        self.field_type = field_type
        self.primary_key = primary_key
        self.name = name
        self.default = default

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        return f"<Field {self.name!r}>"

    def __str__(self):
        text = f"{self.name!r} of type {self.field_type.__name__}"
        if self.primary_key:
            text += " (primary key)"

        return text

    @property
    def set_by_database(self):
        """This field is to be set by the database only."""
        return self.field_type is int and self.primary_key

    @property
    def has_default(self):
        """Has this field got a default value?"""
        return self.default is not _NOT_SET

    def accept(self, value: Any) -> bool:
        """Return whether this value is accepted."""
        accepted = self.field_type
        if accepted in (int, float, str, bytes):
            return isinstance(value, accepted)

        return True



# Copyright (c) 2021, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


"""Module containing the base class BaseCommand, described below."""

from argparse import ArgumentParser

class BaseCommand:

    """Base class of a command-line tool.

    This class is mainly a bridge between a set of commands and
    the 'argparse' module.  The latter manage sub-commands, but
    the automation of function calls isn't very straightforward.
    This wrapper is used to create a simple (and as clean as
    possible) distinction between sub-commands.

    To create a new command, you should inherit from this class.
    The '__init__' method can be redefined to add arguments to the
    parser (the 'parser' attribute).  The 'execute' method is
    also to be redefined.

    """

    def __init__(self, parser=None):
        parser = parser or ArgumentParser()
        self.parser = parser
        self.subparsers = None
        self.parser.set_defaults(func=self.execute)

    def add_subcommand(self, CommandClass):
        """Add a sub-command."""
        if self.subparsers is None:
            self.subparsers = self.parser.add_subparsers()
        parser = self.subparsers.add_parser(CommandClass.name)
        command = CommandClass(parser)

    def execute(self, args):
        """Execute the command."""
        raise NotImplementedError



# Copyright (c) 2021, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test the model API."""

from test.base import BaseTest

from mustang import Field, Model
from mustang.exceptions import *

class Person(Model):

    """A simple person."""

    first = Field(str)
    last = Field(str)
    age = Field(int)
    height = Field(float)

class TestModels(BaseTest):

    """Test the model API."""

    def setUp(self):
        super().setUp()
        self.db.bind((Person, ))

    def test_create(self):
        """Create several instances."""
        # Create a person with all fields.
        person = Person.create(first="Vincent", last="Le Goff", age=31, height=1.72)
        self.assertEqual(person.first, "Vincent")
        self.assertEqual(person.last, "Le Goff")
        self.assertEqual(person.age, 31)
        self.assertEqual(person.height, 1.72)

        # Check that this person has a valid ID.
        self.assertIsNotNone(person.id)

        # Creating a person with missing fields should raise an error.
        with self.assertRaises(MissingArgument):
            Person.create(first="Vincent")

        # Creating a person with an ID shouldn't be allowed.
        with self.assertRaises(ForbiddenArgument):
            Person.create(id=4, first="Vincent", last="Le Goff",
                    age=31, height=1.72)

    def test_get(self):
        """Test to get a model."""
        person = Person.create(first="Vincent", last="Le Goff", age=31,
                height=1.72)
        self.assertIsNotNone(Person.get(id=person.id))
        self.assertIsNone(Person.get(id=person.id + 1))

    def test_update(self):
        """Test to update a model."""
        person = Person.create(first="Vincent", last="Le Goff", age=31,
                height=1.72)
        person.age = 8
        self.assertEqual(person.age, 8)

        # Check that the same result is obtained through getting the object.
        self.assertEqual(Person.get(id=person.id).age, 8)

        # But editing the ID raises an error.
        with self.assertRaises(SetByDatabase):
            person.id = 32

    def test_delete(self):
        """Create and delete a model."""
        person = Person.create(first="Vincent", last="Le Goff", age=31,
                height=1.72)
        self.assertIsNotNone(Person.get(id=person.id))
        person.delete()
        self.assertIsNone(Person.get(id=person.id))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pymustang",
    version="0.1",
    author="Vincent Le Goff",
    author_email="vincent.legoff.srs@gmail.com",
    description="A lightweight, Sqlite ORM in Python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vincent-lg/mustang/",
    packages=setuptools.find_packages(),
    python_requires=">=3.6",
    install_requires = [],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    entry_points = {
        'console_scripts': ['mustang=mustang.commands.main:main'],
    },
)



# Copyright (c) 2021, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Module containing Mustang's exceptions."""

class MustangException(RuntimeError):

    """Exceptions specific to Mustang."""

    pass

class InvalidArgument(MustangException):

    """
    Model creation failed because of arguments.

    This exception is raised when the user wants to create a model
    but a specified (or missing) argument is present.

    """

    def __init__(self, model, field):
        self.model = model.__name__
        self.argument = field.name

class ForbiddenArgument(InvalidArgument):

    """Exception raised when this argument shouldn't be set."""

    def __str__(self):
        return (
                f"model {self.model}: can't set {self.field!r}, "
                "let the database do it"
        )

class MissingArgument(InvalidArgument):

    """Exception raised when the specified field is missing."""

    def __str__(self):
        return f"model {self.model}: missing the argument {self.argument!r}"

class SetByDatabase(InvalidArgument):

    """Exception raised when updating a field only the database should set."""

    def __str__(self):
        return (
                f"model {self.model}: only the database should set "
                f"{self.argument!r}"
        )



# Copyright (c) 2021, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Base class for all models."""

from typing import Dict, Optional, Type

from mustang.exceptions import SetByDatabase
from mustang.schema.field import Field
from mustang.schema.schema import ModelSchema, Operation

MODELS = set()
_NOT_SET = object()

class MetaModel(type):

    """Metaclass for all models."""

    def __new__(cls, name, bases, attrs):
        cls = super().__new__(cls, name, bases, attrs)
        if cls.__name__ != "Model":
            MODELS.add(cls)
            cls._fields = MetaModel.get_fields(cls)
        return cls

    @staticmethod
    def get_fields(model: Type["Model"]) -> Dict[str, Field]:
        """Get the tuple of fields from a model."""
        fields = {}
        for key, value in tuple(model.__dict__.items()):
            if isinstance(value, Field):
                value.name = key
                fields[key] = value

        # If there is no PrimaryKey field, add one.
        if not any(field for field in fields.values() if field.primary_key):
            if any(field for field in fields.values() if field.name == "id"):
                raise ValueError(
                        f"model {model!r}: no primary key is defined, but "
                        "there already exists a field of name 'id', so no "
                        "key can be added.  You should specify ONE "
                        "primary key in your model class"
                )

            primary_key = Field(int, primary_key=True, name="id")
            fields = dict(id=primary_key, **fields)
        elif len(field for field in fields.values() if field.primary_key) < 1:
            raise ValueError(
                    f"model {model!r}: there are at least two primary key "
                    "fields, which is not allowed.  Please choose one "
                    "primary key field, although you can have secondary "
                    "key fields, and composite keys, if you want"
            )

        return fields

    def __str__(self):
        text = f"{self.__name__} ("
        if self._fields:
            indent = "\n" + " " * 4
            text += indent + indent.join(
                    str(field) for field in self._fields.values())
            text += "\n)"
        else:
            text += ")"

        return text

    def load_schema(self):
        """Load the model's schema."""
        self._schema = ModelSchema(self, self._fields)

    def create(self, *args, **kwargs):
        """Create and return a model instance."""
        fields = self._schema.extract(args, kwargs, Operation.CREATION)
        return self._engine.create_instance(self, fields)

    def get(self, *args, **kwargs):
        """
        Get a model instance from the database.

        Optional and keyword arguments are supported.

        Returns:
            instance (Model): the model instance or None.

        Note:
            If no model is found, or if several models with these fields
            are found, return None.

        """
        fields = self._schema.extract(args, kwargs, operation=Operation.PORTION)
        return self._engine.get_instance(self, fields)


class Model(metaclass=MetaModel):

    """
    Base class for all database models.

    Models match tables in a database.  A class in-code contains fields,
    each field being a database column when stored.  Models are not
    directly linked to a database, though once instantiated, models
    usually can connect to a database engine to perform operations.
    A database can decide to map all models (all imported classes inheriting
    from `Model`) or a specific set.

    """

    _alt_name: Optional[str] = None
    _engine: Optional['mustang.engine.base.BaseEngine'] = None
    _fields = {}

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)
        self._has_init = True

    def __repr__(self):
        pk = self._schema.primary_key
        value = getattr(self, pk.name)
        return f"<{type(self).__name__}({pk.name}={value!r})>"

    def __setattr__(self, key, value):
        """If a field is updated, notify the database engine."""
        field = self._fields.get(key, _NOT_SET)
        has_init = getattr(self, "_has_init", False)
        if has_init and field and not key.startswith("_"):
            if field.set_by_database:
                raise SetByDatabase(type(self), field)

            field.accept(value)
            self._engine.update_instance(self, field, value)
        else:
            super().__setattr__(key, value)

    def delete(self):
        """Remove this object from the database."""
        return self._engine.delete_instance(self)



# Copyright (c) 2021, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


"""Module containing the CheckCommand class, described below."""

from mustang.commands.base import BaseCommand

class CheckCommand(BaseCommand):

    """Commands 'mustang check'.

    This command is used to check the current database schema and
    see the modifications that are needed, including the migrations.

    """

    name = "check"

    def __init__(self, parser=None):
        BaseCommand.__init__(self, parser)

    def execute(self, args):
        """Execute the command."""
        print("Reached check!")
